"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ContentSearchService = void 0;
const tslib_1 = require("tslib");
const di_1 = require("@opensumi/di");
const ide_connection_1 = require("@opensumi/ide-connection");
const ide_core_node_1 = require("@opensumi/ide-core-node");
const node_1 = require("@opensumi/ide-logs/lib/node");
const ide_process_1 = require("@opensumi/ide-process");
const vscode_ripgrep_1 = require("@opensumi/vscode-ripgrep");
const common_1 = require("../common");
const { replaceAsarInPath } = ide_core_node_1.path;
/**
 * Convert the length of a range in `text` expressed in bytes to a number of
 * characters (or more precisely, code points).  The range starts at character
 * `charStart` in `text`.
 */
function byteRangeLengthToCharacterLength(text, charStart, byteLength) {
    let char = charStart;
    for (let byteIdx = 0; byteIdx < byteLength; char++) {
        const codePoint = text.charCodeAt(char);
        if (codePoint < 0x7f) {
            byteIdx++;
        }
        else if (codePoint < 0x7ff) {
            byteIdx += 2;
        }
        else if (codePoint < 0xffff) {
            byteIdx += 3;
        }
        else if (codePoint < 0x10ffff) {
            byteIdx += 4;
        }
        else {
            throw new Error('Invalid UTF-8 string');
        }
    }
    return char - charStart;
}
let ContentSearchService = class ContentSearchService extends ide_connection_1.RPCService {
    constructor() {
        super();
        this.processMap = new Map();
        this.logger = this.loggerManager.getLogger(node_1.SupportLogNamespace.Node);
    }
    searchStart(searchId, searchProcess) {
        this.sendResultToClient([], searchId, common_1.SEARCH_STATE.doing);
        this.processMap.set(searchId, searchProcess);
    }
    searchEnd(searchId) {
        this.sendResultToClient([], searchId, common_1.SEARCH_STATE.done);
        this.processMap.delete(searchId);
    }
    searchError(searchId, error) {
        this.sendResultToClient([], searchId, common_1.SEARCH_STATE.error, error);
        this.processMap.delete(searchId);
    }
    async search(searchId, what, rootUris, opts) {
        const args = this.getSearchArgs(opts);
        if (opts && opts.matchWholeWord && !opts.useRegExp) {
            what = what.replace(/[-\\{}*+?|^$.[\]()#]/g, '\\$&');
            if (!/\B/.test(what.charAt(0))) {
                what = '\\b' + what;
            }
            if (!/\B/.test(what.charAt(what.length - 1))) {
                what = what + '\\b';
            }
        }
        const searchInfo = {
            searchId,
            resultLength: 0,
            dataBuf: '',
        };
        const processOptions = {
            command: replaceAsarInPath(vscode_ripgrep_1.rgPath),
            args: [...args, what].concat(rootUris.map((root) => ide_core_node_1.FileUri.fsPath(root))),
        };
        const rgProcess = this.processFactory.create(processOptions);
        this.searchStart(searchInfo.searchId, rgProcess);
        rgProcess.onError((error) => {
            let errorCode = error.code;
            // Try to provide somewhat clearer error messages, if possible.
            if (errorCode === 'ENOENT') {
                errorCode = 'could not find the ripgrep (rg) binary';
            }
            else if (errorCode === 'EACCES') {
                errorCode = 'could not execute the ripgrep (rg) binary';
            }
            const errorStr = `An error happened while searching (${errorCode}).`;
            this.logger.error(errorStr);
            this.searchError(searchInfo.searchId, errorStr);
        });
        rgProcess.outputStream.on('data', (chunk) => {
            searchInfo.dataBuf = searchInfo.dataBuf + chunk;
            this.parseDataBuffer(searchInfo, opts, rootUris);
        });
        rgProcess.onExit(() => {
            this.searchEnd(searchInfo.searchId);
        });
        return searchInfo.searchId;
    }
    cancel(searchId) {
        const process = this.processMap.get(searchId);
        if (process) {
            process.dispose();
            this.processMap.delete(searchId);
        }
        return Promise.resolve();
    }
    parseDataBuffer(searchInfo, opts, rootUris) {
        const lines = searchInfo.dataBuf.toString().split('\n');
        const result = [];
        let filterFileWithGlobRelativePath;
        if (lines.length < 1) {
            return;
        }
        if (rootUris && opts) {
            filterFileWithGlobRelativePath = new common_1.FilterFileWithGlobRelativePath(rootUris, opts.include || []);
        }
        lines.some((line) => {
            // 读一行清理一行
            const eolIdx = searchInfo.dataBuf.indexOf('\n');
            if (eolIdx > -1) {
                searchInfo.dataBuf = searchInfo.dataBuf.slice(eolIdx + 1);
            }
            let lintObj;
            try {
                lintObj = JSON.parse(line.trim());
            }
            catch (e) { }
            if (!lintObj) {
                return;
            }
            if (lintObj.type === 'match') {
                const data = lintObj.data;
                const file = data.path.text;
                const line = data.line_number;
                const lineText = data.lines.text;
                if (file === undefined || lineText === undefined) {
                    return;
                }
                for (const submatch of data.submatches) {
                    const startByte = submatch.start;
                    const endByte = submatch.end;
                    const character = byteRangeLengthToCharacterLength(lineText, 0, startByte);
                    const matchLength = byteRangeLengthToCharacterLength(lineText, character, endByte - startByte);
                    const fileUri = ide_core_node_1.FileUri.create(file);
                    const fileUriSting = fileUri.toString();
                    if (filterFileWithGlobRelativePath && !filterFileWithGlobRelativePath.test(fileUriSting)) {
                        continue;
                    }
                    const searchResult = (0, common_1.cutShortSearchResult)({
                        fileUri: fileUriSting,
                        line,
                        matchStart: character + 1,
                        matchLength,
                        lineText: lineText.replace(/[\r\n]+$/, ''),
                    });
                    if (opts && opts.maxResults && searchInfo.resultLength >= opts.maxResults) {
                        // 达到设置上限，停止搜索
                        this.logger.debug('Reached the set upper limit, stop searching.');
                        this.cancel(searchInfo.searchId);
                        return true;
                    }
                    result.push(searchResult);
                    searchInfo.resultLength++;
                }
            }
        });
        if (!result || result.length === 0) {
            return;
        }
        this.sendResultToClient(result, searchInfo.searchId);
    }
    sendResultToClient(data, id, searchState, error) {
        if (this.client) {
            this.client.onSearchResult({
                data,
                id,
                searchState,
                error,
            });
        }
    }
    getSearchArgs(options) {
        const args = ['--json', '--max-count=100'];
        args.push(options && options.matchCase ? '--case-sensitive' : '--ignore-case');
        if (options && options.includeIgnored) {
            args.push('-uu');
        }
        if (options && options.include) {
            for (const include of options.include) {
                if (include !== '') {
                    args.push('--glob=' + (0, common_1.anchorGlob)(include));
                }
            }
        }
        if (options && options.exclude) {
            for (const exclude of options.exclude) {
                if (exclude !== '') {
                    args.push('--glob=!' + (0, common_1.anchorGlob)(exclude));
                }
            }
        }
        if (options && options.encoding && options.encoding !== 'utf8') {
            args.push('--encoding', options.encoding);
        }
        if ((options && options.useRegExp) || (options && options.matchWholeWord)) {
            args.push('--regexp');
        }
        else {
            args.push('--fixed-strings');
            args.push('--');
        }
        return args;
    }
    dispose() {
        this.processMap.forEach((v) => {
            v.dispose();
        });
    }
};
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_process_1.IProcessFactory),
    tslib_1.__metadata("design:type", Object)
], ContentSearchService.prototype, "processFactory", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(node_1.ILogServiceManager),
    tslib_1.__metadata("design:type", Object)
], ContentSearchService.prototype, "loggerManager", void 0);
ContentSearchService = tslib_1.__decorate([
    (0, di_1.Injectable)(),
    tslib_1.__metadata("design:paramtypes", [])
], ContentSearchService);
exports.ContentSearchService = ContentSearchService;
//# sourceMappingURL=content-search.service.js.map